/* ide-threading.h
 *
 * Copyright 2018-2019 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <libide-core.h>

#define IDE_THREADING_INSIDE

#include "ide-environment.h"
#include "ide-environment-variable.h"
#include "ide-subprocess-launcher.h"
#include "ide-subprocess-supervisor.h"
#include "ide-subprocess.h"
#include "ide-task.h"
#include "ide-thread-pool.h"

#undef IDE_THREADING_INSIDE
