@binding-set builder-command-bar-entry
{
  bind "<ctrl>a" { "move-cursor" (paragraph-ends, -1, 0) };
  bind "<ctrl>e" { "move-cursor" (paragraph-ends, 1, 0) };

  bind "<ctrl>u" { "move-cursor" (paragraph-ends, -1, 0)
                   "delete-from-cursor" (paragraph-ends, 1) };
  bind "<ctrl>h" { "delete-from-cursor" (chars, -1) };
  bind "<ctrl>w" { "delete-from-cursor" (word-ends, -1) };

}

@binding-set builder-workbench-bindings
{
  bind "<ctrl>comma" { "action" ("app", "preferences", "") };
}

entry.gb-command-bar-entry {
  -gtk-key-bindings: builder-command-bar-entry;
}

window.workbench {
  -gtk-key-bindings: builder-workbench-bindings;
}
