/* gbp-glade-private.h
 *
 * Copyright 2018-2019 Christian Hergert <chergert@redhat.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <libide-gui.h>
#include <gladeui/glade.h>
#include <gladeui/glade-adaptor-chooser.h>

#include "gbp-glade-page.h"

G_BEGIN_DECLS

struct _GbpGladePage
{
  IdePage              parent_instance;

  GFile               *file;
  GladeProject        *project;
  DzlSignalGroup      *project_signals;

  GladeDesignView     *designer;
  GladeAdaptorChooser *chooser;
  GtkBox              *main_box;
};

void     _gbp_glade_page_init_actions   (GbpGladePage  *self);
void     _gbp_glade_page_init_shortcuts (GtkWidget     *widget);
void     _gbp_glade_page_update_actions (GbpGladePage  *self);
gboolean _gbp_glade_page_reload         (GbpGladePage  *self);
gboolean _gbp_glade_page_save           (GbpGladePage  *self,
                                         GError       **error);

G_END_DECLS
